//! This is a skeleton Radicle CI adapter for demonstration purposes.
//!
//! This program shows how to build a very simple adapter for Radicle
//! CI. As per the CI broker/adapter protocol, it reads a trigger
//! request message from its stdin, and writes two response messages
//! to its stdout. Based on the trigger, it clones the relevant Git
//! repository, and checks out the relevant Git commit, then runs
//! executed a silly CI plan. To keep things simple, and quick, in
//! this program the plan is "does the .radicle directory exist?". If
//! it exists, CI succeeds, otherwise it fails. If any of the previous
//! steps fails, CI also fails.
//!
//! Any errors are reported to stderr. If things go wrong very early,
//! the first response message may be a failure message, rather than a
//! "triggered" message. The CI broke will hopefully handle this.

use std::{
    error::Error,
    fs::create_dir_all,
    path::{Path, PathBuf},
    process::Command,
};

use clap::Parser;
use html_page::{Element, HtmlPage, Tag};
use tempfile::tempdir;

use radicle::prelude::{Profile, RepoId};
use radicle_ci_broker::msg::{
    helper::{read_request, write_failed, write_succeeded, write_triggered},
    Request, RunId,
};
use radicle_git_ext::Oid;

use radicle_ci_ambient::{adminlog::AdminLog, config::Config};

// Exit code to indicate we didn't get one from the process.
const NO_EXIT: i32 = 999;

// Name of environment variable that has filename for configuration
// file.
const CONFIG_ENV: &str = "RADICLE_CI_AMBIENT";

fn main() -> Result<(), AdapterError> {
    match fallible_main() {
        Ok(true) => write_succeeded().map_err(AdapterError::Succeded)?,
        Ok(false) => write_failed().map_err(AdapterError::Failed)?,
        Err(err) => {
            eprintln!("ERROR: {err}");
            write_failed().map_err(AdapterError::Failed)?;
            std::process::exit(1);
        }
    }

    Ok(())
}

fn fallible_main() -> Result<bool, AdapterError> {
    let args = Args::parse();

    let config = Config::load_via_env(CONFIG_ENV)?;
    let mut adminlog = config.open_log()?;
    adminlog.writeln("radicle-ci-ambient starts")?;

    // Read the trigger request message and extract the repository ID
    // and commit from it, if possible.
    let req = read_request().map_err(AdapterError::ReadRequest)?;
    let repoid = req.repo();
    let commit = req.commit().map_err(AdapterError::Commit)?;

    // Create a temporary directory and a sub-directory `src` where
    // we'll put the source code from Git.
    let tmp = tempdir().map_err(AdapterError::TempDir)?;
    let src = tmp.path().join("src");
    get_sources(&mut adminlog, args.dry_run, repoid, commit, &src)?;

    // Invent an adapter run ID and tell it to the CI broker.
    let run_id = RunId::default();
    let project_name = repoid.to_string();
    let project_log_dir = config.logdir.join(&project_name);
    if !project_log_dir.exists() {
        create_dir_all(&project_log_dir)
            .map_err(|err| AdapterError::CreateDirs(project_log_dir, err))?;
    }
    let run_log_relative = format!("{project_name}/{run_id}.html");

    let url = config
        .base_url
        .clone()
        .map(|base| format!("{base}/{run_log_relative}"));
    write_triggered(&run_id, url.as_deref()).map_err(AdapterError::Triggered)?;

    // Execute CI in the source directory.
    let result = execute_ci(
        &mut adminlog,
        args.dry_run,
        &config,
        &req,
        &project_name,
        &run_log_relative,
        &src,
    );
    adminlog.writeln(&format!(
        "radicle-ci-ambient ends: Ambient result is {result:?}"
    ))?;

    if let Err(err) = &result {
        adminlog.writeln(&format!("ERROR: {err}"))?;
        let mut err = err.source();
        while let Some(cause) = err {
            adminlog.writeln(&format!("caused by {cause}"))?;
            err = cause.source();
        }
    }

    result
}

#[derive(Parser)]
#[clap(version)]
struct Args {
    #[clap(long)]
    dry_run: bool,
}

// Get sources from the local node.
fn get_sources(
    adminlog: &mut AdminLog,
    dry_run: bool,
    repoid: RepoId,
    commit: Oid,
    src: &Path,
) -> Result<(), AdapterError> {
    adminlog.writeln("load Radicle profile")?;
    let profile = Profile::load().map_err(AdapterError::Profile)?;
    let storage = profile.storage.path();
    let repo_path = storage.join(repoid.canonical());

    git_clone(adminlog, dry_run, &repo_path, src)?;
    git_checkout(adminlog, dry_run, commit, src)?;

    Ok(())
}

// Run `git clone` for the repository.
fn git_clone(
    adminlog: &mut AdminLog,
    dry_run: bool,
    repo_path: &Path,
    src: &Path,
) -> Result<(), AdapterError> {
    runcmd(
        adminlog,
        dry_run,
        &[
            "git",
            "clone",
            repo_path.to_str().unwrap(),
            src.to_str().unwrap(),
        ],
        Path::new("."),
    )?;
    Ok(())
}

// Check out the requested commit.
fn git_checkout(
    adminlog: &mut AdminLog,
    dry_run: bool,
    commit: Oid,
    src: &Path,
) -> Result<(), AdapterError> {
    runcmd(
        adminlog,
        dry_run,
        &["git", "config", "advice.detachedHead", "false"],
        src,
    )?;
    runcmd(
        adminlog,
        dry_run,
        &["git", "checkout", &commit.to_string()],
        src,
    )?;
    Ok(())
}

// Run a program.
fn runcmd(
    adminlog: &mut AdminLog,
    dry_run: bool,
    argv: &[&str],
    cwd: &Path,
) -> Result<(i32, Vec<u8>), AdapterError> {
    assert!(!argv.is_empty());

    if dry_run {
        adminlog.writeln(&format!("runcmd: pretend to run: argv={argv:?}"))?;
        return Ok((0, vec![]));
    }

    adminlog.writeln(&format!("runcmd: argv={argv:?}"))?;
    let output = Command::new("bash")
        .arg("-c")
        .arg(r#""$@" 2>&1"#)
        .arg("--")
        .args(argv)
        .current_dir(cwd)
        .output()
        .map_err(|err| AdapterError::Invoke("bash", err))?;

    let exit = output.status.code().unwrap_or(NO_EXIT);

    if exit != 0 {
        adminlog.writeln(&format!("runcmd: exit={exit}"))?;
        indented(adminlog, "stdout", &output.stdout);
        indented(adminlog, "stderr", &output.stderr);
    }

    Ok((exit, output.stdout))
}

fn indented(adminlog: &mut AdminLog, msg: &str, bytes: &[u8]) {
    if !bytes.is_empty() {
        adminlog.writeln(&format!("{msg}:")).ok();
        let text = String::from_utf8_lossy(bytes);
        for line in text.lines() {
            adminlog.writeln(&format!("    {line}")).ok();
        }
    }
}

// This is what executes the steps CI is meant to execute. Edit this
// according to your needs.
fn execute_ci(
    adminlog: &mut AdminLog,
    dry_run: bool,
    config: &Config,
    trigger: &Request,
    project_name: &str,
    run_log_relative: &str,
    src: &Path,
) -> Result<bool, AdapterError> {
    adminlog.writeln(&format!("execute_ci: src={}", src.display()))?;

    let tmp = tempdir().map_err(AdapterError::TempDir)?;
    let projects_filename = tmp.path().join("projects.yaml");

    let filename = src.join(".radicle/ambient.yaml");
    let distilled = if dry_run {
        ambient::DistilledPlan::default()
    } else {
        ambient::DistilledPlan::read(&filename)?
    };

    let projects = ambient::ProjectsBuilder::default()
        .name(project_name)
        .source(src)
        .image(&config.image)
        .artifact_max_size(1024 * 1024 * 1024)
        .cache_max_size(20 * 1024 * 1024 * 1024)
        .plan(&distilled)
        .build();

    let projects = serde_yml::to_string(&projects).map_err(AdapterError::ToYaml)?;
    indented(adminlog, "projects", projects.as_bytes());

    std::fs::write(&projects_filename, projects.as_bytes()).map_err(AdapterError::WriteProjects)?;

    adminlog.writeln("execute_ci: run ambient-driver")?;
    let filename = projects_filename.display().to_string();
    let (exit, _) = runcmd(
        adminlog,
        dry_run,
        &[
            "env",
            "AMBIENT_LOG=debug",
            "ambient-driver",
            "run",
            "--force",
            "--projects",
            filename.as_str(),
        ],
        Path::new("."),
    )?;

    let (exit2, run_log_raw) = runcmd(
        adminlog,
        dry_run,
        &["ambient-driver", "log", project_name],
        Path::new("."),
    )?;
    if exit2 == 0 {
        if !config.logdir.exists() {
            adminlog.writeln("create log directory")?;
            std::fs::create_dir_all(&config.logdir)
                .map_err(|err| AdapterError::CreateDirs(config.logdir.clone(), err))?;
        }

        adminlog.writeln("write run log")?;
        let filename = config.logdir.join(run_log_relative).with_extension("bin");
        std::fs::write(&filename, &run_log_raw)
            .map_err(|err| AdapterError::WriteRunLog(filename.clone(), err))?;

        adminlog.writeln("convert terminal output to HTML and write that")?;
        let page = html_log(trigger, project_name, &run_log_raw)?;
        let filename = config.logdir.join(run_log_relative).with_extension("html");
        std::fs::write(&filename, page.to_string())
            .map_err(|err| AdapterError::WriteRunLog(filename.clone(), err))?;
    }

    if dry_run {
        adminlog.writeln("execute_ci: Ambient is pretending to be OK")?;
        Ok(true)
    } else if exit == 0 {
        adminlog.writeln("execute_ci: Ambient is OK")?;
        Ok(true)
    } else {
        adminlog.writeln(&format!("ERROR: ambient-driver exit code {exit}"))?;
        Ok(false)
    }
}

fn html_log(
    trigger: &Request,
    project_name: &str,
    run_log: &[u8],
) -> Result<HtmlPage, AdapterError> {
    let title = format!("CI run for {project_name}");
    let trigger = serde_json::to_string_pretty(trigger).map_err(AdapterError::ToJson)?;
    let run_log = String::from_utf8_lossy(run_log);

    Ok(HtmlPage::default()
        .with_head_element(Element::new(Tag::Title).with_text(&title))
        .with_body_element(Element::new(Tag::H1).with_text(&title))
        .with_body_element(Element::new(Tag::H2).with_text("Triggered by"))
        .with_body_element(Element::new(Tag::Pre).with_text(&trigger))
        .with_body_element(Element::new(Tag::H2).with_text("Run log"))
        .with_body_element(Element::new(Tag::Pre).with_text(&run_log)))
}

#[allow(dead_code)]
mod ambient {
    use std::{
        collections::HashMap,
        path::{Path, PathBuf},
    };

    use ambient_driver::action::{TrustedAction, UnsafeAction};
    use serde::{Deserialize, Serialize};

    use super::AdapterError;

    // Read `.radicle/ambient.yaml` from the source tree.
    #[derive(Debug, Default, Deserialize)]
    #[serde(deny_unknown_fields)]
    pub struct DistilledPlan {
        pre_plan: Option<Vec<TrustedAction>>,
        plan: Option<Vec<UnsafeAction>>,
        post_plan: Option<Vec<TrustedAction>>,
    }

    impl DistilledPlan {
        pub fn read(filename: &Path) -> Result<Self, AdapterError> {
            let data = std::fs::read(filename)
                .map_err(|err| AdapterError::ReadPlan(filename.into(), err))?;
            serde_yml::from_slice(&data)
                .map_err(|err| AdapterError::ParsePlan(filename.into(), err))
        }
    }

    #[derive(Debug, Serialize)]
    pub struct Project {
        source: PathBuf,
        image: PathBuf,
        pre_plan: Vec<TrustedAction>,
        plan: Vec<UnsafeAction>,
        post_plan: Vec<TrustedAction>,
        artifact_max_size: Option<u64>,
        cache_max_size: Option<u64>,
    }

    #[derive(Debug, Serialize)]
    pub struct Projects {
        projects: HashMap<String, Project>,
    }

    impl Projects {
        pub fn write(&self, filename: &Path) -> Result<(), AdapterError> {
            let text = serde_yml::to_string(&self).map_err(AdapterError::ToYaml)?;
            std::fs::write(filename, text.as_bytes()).map_err(AdapterError::WriteProjects)?;
            Ok(())
        }
    }

    #[derive(Debug, Default)]
    pub struct ProjectsBuilder {
        name: Option<String>,
        source: Option<PathBuf>,
        image: Option<PathBuf>,
        pre_plan: Vec<TrustedAction>,
        plan: Vec<UnsafeAction>,
        post_plan: Vec<TrustedAction>,
        artifact_max_size: Option<u64>,
        cache_max_size: Option<u64>,
    }

    impl ProjectsBuilder {
        pub fn name(mut self, name: &str) -> Self {
            self.name = Some(name.into());
            self
        }

        pub fn source(mut self, path: &Path) -> Self {
            self.source = Some(path.into());
            self
        }

        pub fn image(mut self, path: &Path) -> Self {
            self.image = Some(path.into());
            self
        }

        pub fn artifact_max_size(mut self, size: u64) -> Self {
            self.artifact_max_size = Some(size);
            self
        }

        pub fn cache_max_size(mut self, size: u64) -> Self {
            self.cache_max_size = Some(size);
            self
        }

        pub fn plan(mut self, plan: &DistilledPlan) -> Self {
            if let Some(actions) = &plan.pre_plan {
                self.pre_plan = actions.to_vec();
            }
            if let Some(actions) = &plan.plan {
                self.plan = actions.to_vec();
            }
            if let Some(actions) = &plan.post_plan {
                self.post_plan = actions.to_vec();
            }
            self
        }

        pub fn build(self) -> Projects {
            let projects: HashMap<String, Project> = HashMap::from([(
                self.name.unwrap(),
                Project {
                    source: self.source.expect("ProjectsBuilder::source is set"),
                    image: self.image.expect("ProjectsBuilder::image is set"),
                    pre_plan: self.pre_plan.clone(),
                    plan: self.plan.clone(),
                    post_plan: self.post_plan.clone(),
                    artifact_max_size: self.artifact_max_size,
                    cache_max_size: self.cache_max_size,
                },
            )]);
            Projects { projects }
        }
    }
}

#[derive(Debug, thiserror::Error)]
enum AdapterError {
    #[error("failed to write message reporting run has been triggered")]
    Triggered(#[source] radicle_ci_broker::msg::helper::MessageHelperError),

    #[error("failed to write response message reporting success")]
    Succeded(#[source] radicle_ci_broker::msg::helper::MessageHelperError),

    #[error("failed to write response message reporting failure")]
    Failed(#[source] radicle_ci_broker::msg::helper::MessageHelperError),

    #[error("failed to read request from stdin")]
    ReadRequest(#[source] radicle_ci_broker::msg::helper::MessageHelperError),

    #[error(transparent)]
    Config(#[from] radicle_ci_ambient::config::ConfigError),

    #[error(transparent)]
    Log(#[from] radicle_ci_ambient::adminlog::LogError),

    #[error("failed to extract commit from request message")]
    Commit(#[source] radicle_ci_broker::msg::MessageError),

    #[error("failed to create a temporary directory")]
    TempDir(#[source] std::io::Error),

    #[error("failed to load Radicle profile")]
    Profile(#[source] radicle::profile::Error),

    #[error("failed to invoke command {0}")]
    Invoke(&'static str, #[source] std::io::Error),

    #[error("failed to serialize Ambient project list to YAML")]
    ToYaml(#[source] serde_yml::Error),

    #[error("failed to write Ambient project list to file")]
    WriteProjects(#[source] std::io::Error),

    #[error("failed to create directory {0}")]
    CreateDirs(PathBuf, #[source] std::io::Error),

    #[error("failed to write run log to file")]
    WriteRunLog(PathBuf, #[source] std::io::Error),

    #[error("failed to read CI plan from project {0}")]
    ReadPlan(PathBuf, #[source] std::io::Error),

    #[error("failed to parse CI plan as YAML fro file {0}")]
    ParsePlan(PathBuf, #[source] serde_yml::Error),

    #[error("failed to serialize trigger message as JSON")]
    ToJson(#[source] serde_json::Error),
}
